// cppvm 'collection'
#include "cppvm.h"

// return a collection member value as double
double
valueAsDouble(cppvmObject* obj)
{
  int msgTag;
  msgTag = obj->cppvmMsgTag();
  if(msgTag != CPPvmMsgTagClassString &&
     msgTag != CPPvmMsgTagClassBool   &&     
     msgTag != CPPvmMsgTagClassInt    &&     
     msgTag != CPPvmMsgTagClassLong   &&     
     msgTag != CPPvmMsgTagClassFloat  &&     
     msgTag != CPPvmMsgTagClassDouble &&     
     msgTag != CPPvmMsgTagClassChar){
    // not a class derived from 'cppvmSingleData'
    // return dummy value
    return 0.0;
  }
  else{
    // read the value of a 'cppvmSingleData' type object as double
    return ((cppvmSingleData*)obj)->readDouble();
  }
}

// compare function (descending) for two members of a collection
bool 
compareDescending(cppvmObject* obj1, cppvmObject* obj2){
  double val1, val2;

  val1 = valueAsDouble(obj1);
  val2 = valueAsDouble(obj2);

  if(val1<val2){
    return true;
  }
  else{
    return false;
  }
}

// compare function (ascending) for two members of a collection
bool 
compareAscending(cppvmObject* obj1, cppvmObject* obj2){
  return !(compareDescending(obj1, obj2));
}

// indicate the class of an 'cppvmObject'
void
objectClass(cppvmObject& obj, int msgTag, cppvmString& classString){
  if(msgTag == CPPvmMsgTagClassString){
    classString = "cppvmString";
  }
  else if(msgTag == CPPvmMsgTagClassBool){
    classString = "cppvmBool";
  }
  else if(msgTag == CPPvmMsgTagClassInt){
    classString = "cppvmInt";
  }
  else if(msgTag == CPPvmMsgTagClassLong){
    classString = "cppvmLong";
  }
  else if(msgTag == CPPvmMsgTagClassFloat){
    classString = "cppvmFloat";
  }
  else if(msgTag == CPPvmMsgTagClassDouble){
    classString = "cppvmDouble";
  }
  else if(msgTag == CPPvmMsgTagClassChar){
    classString = "cppvmChar";
  }
  else{
    classString = "Unknown";
  }
}

int
main()
{
  // connect to pvm
  cppvmConnection pvmConn;

  // direct all couts to stdout
  cppvmCatchout catchOut(CPPvmStdout);
  // spawn child process 'coll2'
  cppvmSpawnConnection child("coll2");

  unsigned int i;
  
  //////////////////////////////////
  // standard collection handling //
  // with sort                    //
  //////////////////////////////////


  // collection for objects of 
  // 'cppvmObject' and derived classes
  cppvmObjectCollection myColl;

  // add some objects to the collection
  // the objects are now administrated by the collection
  // use the collection methods to modify/delete the objects
  myColl.add(new cppvmFloat(4.1));
  myColl.add(new cppvmInt(4));
  myColl.add(new cppvmBool(false));
  myColl.add(new cppvmString("4.3"));
  myColl.add(new cppvmLong(5));
  myColl.add(new cppvmDouble(4.2));

  cout << "Collection before sort" << endl
       << myColl << endl;

  // a compare function which can compare all members
  // of the collection can be used to sort the collection
  // the compare function must be of type
  // 'bool myCompare(cppvmObject* obj1, cppvmObject* obj2)'
  myColl.sort(compareDescending);
  cout << "Collection after sort descending" << endl
       << myColl << endl;

  // sort again
  myColl.sort(compareAscending);
  cout << "Collection after sort ascending" << endl
       << myColl << endl;

  for(i=0; i<myColl.numElems();i++){
    cppvmString objClass("");

    // the collection knows only objects of class 'cppvmObject'
    // if you use classes derived form 'cppvmObject',
    // you have to downcast them to their real classes 
    // to have access to all methods
    // the unique (well they should be :-) pvm message tags
    // can be used to distinguish between the classes
    
    // get the class of the object as string
    objectClass(myColl[i], myColl[i].cppvmMsgTag(), objClass);

    cout << "collection member number: " << i 
         << " is of class: " << objClass << endl;

    // all objects in this collection are derived from
    // the class 'cppvmSingleData' so methods of this
    // class can be used
    ((cppvmSingleData&)myColl[i]).writeDouble((double) i);

  }
  
  cout << "Collection before sort2" << endl
       << myColl << endl;

  // sort the collection with new values
  myColl.sort(compareDescending);
  cout << "Collection after sort2 descending" << endl
       << myColl << endl;

  // ... and again 
  myColl.sort(compareAscending);
  cout << "Collection after sort2 ascending" << endl
       << myColl << endl;

  
  //////////////////////////////////
  // nested collection            //
  //////////////////////////////////

  // 'cppvmObjectCollection' is also derived from 'cppvmObject',
  // so collections can also be added as members
  // of a collection
  cppvmObjectCollection *myColl2Ptr = new cppvmObjectCollection;
  myColl2Ptr->add(new cppvmFloat(8.8));
  myColl2Ptr->add(new cppvmString("nested collections"));

  cppvmObjectCollection *myColl3Ptr = new cppvmObjectCollection;
  myColl3Ptr->add(new cppvmString("are"));

  // add collection 3 to collection 2
  myColl2Ptr->add(myColl3Ptr);
  // add collection 2 to collection 1
  myColl.add(myColl2Ptr);
  // add new 'cppvmString' to collection 1
  myColl.add(new cppvmString("possible!"));
  cout << "Collection after adding myColl2Ptr" << endl
       << myColl << endl;
  

  //////////////////////////////////
  // send collection              //
  //////////////////////////////////

  // (nested) collections can be sent/received
  // as usual

  // message passing descriptors
  // blocking send to child process
  cppvmSendStream sendStrm(child, CPPvmSBchild);

  // send nested collection
  sendStrm << myColl;

  // remove (delete) all members
  myColl.clean();
  myColl.add(new cppvmString("Hi child! "));

  // send new collection
  sendStrm << myColl;
  
  //////////////////////////////////
  // distributed collection       //
  //////////////////////////////////

  // collections can also be distributed

  // register the object as a distributed object
  myColl.cppvmDistObj("distColl");
  
  for(i=0;i<25000;i++){
    // update the local collection with the value
    // in the message box
    myColl.cppvmReadMboxInfo();
    cout << '\r' << myColl;
    cout.flush();
  }
  cout << endl;
  
  //////////////////////////////////
  // send collection              //
  //////////////////////////////////

  // send collection to show different
  // receive modes on the other side

  myColl.clean();
  myColl.add(new cppvmString("stop it!"));

  sendStrm << myColl << myColl;

  // message passing descriptors
  // blocking receive from child process
  cppvmReceiveStream recStrm(child, CPPvmRBchild);

  recStrm >> myColl;

  cout << "returned from child: " << myColl << endl;

  myColl.remove((unsigned int)0);
  
  // halt pvm
  pvmConn.halt();
  
  return 0;
}
